# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
NEP

Configuration keys:

[core]

google-username = user.name
google-password = secret
google-docs-key = crypticstring

[nep]

# location of archived shipments
backlog-host = example.com
backlog-username = username
backlog-pattern = /path/to/data/*.upd.zip

# location of recent shipments
ftp-host = host.name
ftp-username = username
ftp-password = password
ftp-pattern = some*glob*pattern.zip

# dates to ignore
mute = 2014-01-01, 2014-05-07

"""

from gluish.format import TSV
from gluish.intervals import hourly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.common import Directory, FTPMirror
from siskin.configuration import Config
from siskin.task import DefaultTask
from siskin.utils import copyregions, memoize
import datetime
import glob
import gspread
import json
import luigi
import marcx
import os
import pymarc
import re
import string
import tempfile

config = Config.instance()

class NEPTask(DefaultTask):
    TAG = '003'

    def muted(self):
        """
        Return a set of muted dates.
        """
        mute_value = config.get('nep', 'mute', '')
        try:
            vals = mute_value.split(',')
            return set([datetime.date(*map(int, s.split('-'))) for s in vals])
        except ValueError:
            raise RuntimeError('failed to parse mutes: {0}'.format(mute_value))

    @memoize
    def closest(self):
        if not hasattr(self, 'date'):
            raise RuntimeError('Task has no date: %s' % self)
        task = NEPLatestDate(date=self.date)
        luigi.build([task], local_scheduler=True)
        with task.output().open() as handle:
            date = handle.iter_tsv(cols=('date', 'path')).next().date
        return datetime.date(*(int(v) for v in date.split('-')))

class NEPCopy(NEPTask):

    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    server = luigi.Parameter(default=config.get('nep', 'backlog-host'))
    username = luigi.Parameter(default=config.get('nep', 'backlog-username'))
    pattern = luigi.Parameter(default=config.get('nep', 'backlog-pattern'))

    timeout = luigi.IntParameter(default=15, description='rsnyc IO timeout', significant=False)

    def requires(self):
        return Directory(path=os.path.dirname(self.output().path))

    @timed
    def run(self):
        shellout("""rsync --timeout {timeout} -avz {username}@{server}:{pattern} {directory}""",
                 timeout=self.timeout, username=self.username, server=self.server,
                 pattern=self.pattern, directory=self.input().path)

        with self.output().open('w') as output:
            pattern = os.path.join(self.input().path,
                                   os.path.basename(self.pattern))
            for path in glob.glob(pattern):
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)

class NEPImport(NEPTask):
    """ NEPCopy and FTP. """

    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return {'copy': NEPCopy(),
                'sync': FTPMirror(host=config.get('nep', 'ftp-host'),
                                  username=config.get('nep', 'ftp-username'),
                                  password=config.get('nep', 'ftp-password'),
                                  pattern=config.get('nep', 'ftp-pattern'))}

    @timed
    def run(self):
        target = os.path.dirname(self.output().path)
        for key in ('copy', 'sync'):
            with self.input().get(key).open() as handle:
                for row in handle.iter_tsv(cols=('path',)):
                    basename = os.path.basename(row.path)
                    destination = os.path.join(target, basename)
                    if not os.path.exists(destination):
                        dstdir = os.path.dirname(destination)
                        if not os.path.exists(dstdir):
                            os.makedirs(dstdir)
                        if not os.path.exists(destination):
                            shellout("ln -s {src} {dst}", src=row.path, dst=destination)

        p = os.path.basename(config.get('nep', 'ftp-pattern'))
        pattern = os.path.join(target, p)
        with self.output().open('w') as output:
            for path in sorted(glob.glob(pattern)):
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class NEPDatesAndPaths(NEPTask):
    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return NEPImport()

    @timed
    def run(self):
        """ Parse `nep.ignore`. Extract all dates from the filenames. """
        entries = set()
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                basename = os.path.basename(row.path)
                match = re.search(r'.*_([0-9]{8})_.*', basename)
                if match:
                    date_obj = datetime.datetime.strptime(match.group(1), '%Y%m%d').date()
                    if not date_obj in self.muted():
                        entries.add((date_obj, row.path))
                    else:
                        self.logger.debug('ignoring %s via nep.mute' % date_obj)

        with self.output().open('w') as output:
            for date, path in sorted(entries):
                output.write_tsv(date, path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class NEPLatestDate(NEPTask):
    """ For a given date, return the closest date in the past
    on which EBL shipped. """
    date = luigi.DateParameter(default=datetime.date.today())
    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return NEPDatesAndPaths()

    @timed
    def run(self):
        """ load dates and paths in pathmap, use a custom key
        function to find the closest date, but double check,
        if that date actually lies in the future - if it does, raise
        an exception, otherwise dump the closest date and filename to output """
        pathmap = {}
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('date', 'path')):
                date = datetime.date(*(int(v) for v in row.date.split('-')))
                pathmap[date] = row.path

        def closest_keyfun(date):
            if date > self.date:
                return datetime.timedelta(999999999)
            return abs(date - self.date)

        closest = min(pathmap.keys(), key=closest_keyfun)
        if closest > self.date:
            raise RuntimeError('No shipment before: %s' % min(pathmap.keys()))

        with self.output().open('w') as output:
            output.write_tsv(closest, pathmap.get(closest))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class NEPDates(NEPTask):
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return NEPDatesAndPaths()

    @timed
    def run(self):
        output = shellout("LANG=C awk '{{print $1}}' {input} | LANG=C sort -u > {output}",
                        input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class NEPLatestDateAndPaths(NEPTask):
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return NEPDatesAndPaths()

    @timed
    def run(self):
        with self.input().open() as handle:
            with self.output().open('w') as output:
                for row in handle.iter_tsv(cols=('date', 'path')):
                    date = datetime.date(*(int(v) for v in row.date.split('-')))
                    if date == self.closest():
                        output.write_tsv(*row)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class NEPCombine(NEPTask):
    """ Combine into a single file. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return NEPLatestDateAndPaths(date=self.date)

    @timed
    def run(self):
        _, combined = tempfile.mkstemp(prefix='tasktree-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('date', 'path')):
                shellout("unzip -p {input} >> {output}", input=row.path,
                         output=combined)
        luigi.File(combined).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class NEPTable(NEPTask):
    """ (ID STATUS DATE OFFSET LENGTH) for a single NEP shipments.
    Assumes that a single shipment of NEP will contain no duplicate record IDs. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return NEPCombine(date=self.date)

    @timed
    def run(self):
        mm = shellout("marcmap {input} | LANG=C sort -k1,1 > {output}", input=self.input().path)
        st = shellout("marctotsv {input} 001 @Status {date} | LANG=C sort -k1,1 > {output}", input=self.input().path, date=self.closest())
        output = shellout("paste {status} {marcmap} | cut -f 1,2,3,5,6 > {output}", status=st, marcmap=mm)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class NEPEvents(NEPTask):
    """ (ID STATUS DATE OFFSET LENGTH) for all NEP shipments. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        task = NEPDates()
        luigi.build([task], local_scheduler=True)
        with task.output().open() as handle:
            for row in handle.iter_tsv(cols=('date',)):
                date = datetime.date(*(int(v) for v in row.date.split('-')))
                if date < self.closest():
                    yield NEPTable(date=date)

    @timed
    def run(self):
        _, combined = tempfile.mkstemp(prefix='tasktree-')
        for target in self.input():
            shellout("cat {input} >> {output}", input=target.path,
                     output=combined)
        output = shellout("LANG=C sort -k1,1 -k3,3 {input} > {output}", input=combined)
        luigi.File(output).move(self.output().fn)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class NEPSurface(NEPTask):
    """ Only keep the most recent versions of a record (row). """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return NEPEvents(date=self.date)

    @timed
    def run(self):
        surface = {}
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('id', 'status', 'X', 'X', 'X')):
                if row.status in ('a', 'c', 'n', 'p'):
                    # older entries get overwritten
                    surface[row.id] = row
                elif row.status in ('d',):
                    del surface[row.id]

        _, stopover = tempfile.mkstemp(prefix='tasktree-')
        with luigi.File(path=stopover, format=TSV).open('w') as output:
            for _, row in surface.iteritems():
                output.write_tsv(*row)

        output = shellout("LANG=C sort -k3,3 -k4,4n {input} > {output}",
                          input=stopover)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class NEPSnapshot(NEPTask):
    """ Create the snapshot by utilizing seekmaps/surface. This will create
    a single MARC binary. As of Fall 2013, this file is about 1G in size.
    Non-pandas version.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return NEPSurface(date=self.date)

    @timed
    def run(self):
        output = shellout("cut -f3 {input} | LANG=C sort | LANG=C uniq > {output}", input=self.input().path)
        with open(output) as handle:
            dates = map(string.strip, handle.readlines())

        with self.output().open('w') as output:
            for date in dates:
                task = NEPCombine(date=datetime.date(*(int(v) for v in date.split('-'))))
                luigi.build([task], local_scheduler=True)
                with task.output().open() as fh:
                    seekfile = shellout("""LANG=C grep "{date}" "{input}" | cut -f 4,5 > {output}""", date=str(date), input=self.input().path)
                    with luigi.File(seekfile, format=TSV).open() as handle:
                        seekmap = ((int(offset), int(length)) for offset, length in handle.iter_tsv())
                        copyregions(fh, output, seekmap)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class NEPSubjectFilterCodes(NEPTask):
    """
    Take a file from assets and use as
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        luigi.File(self.assets("nep_subject_filter_codes.tsv")).copy(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class NEPForUBL(NEPTask):
    """
    Filter NEP by bicssc codes. And other filters, relevant for UBL.

        marctotsv -f NA -s '|' .../nep-snapshot/2013-11-20.mrc \
            007 240.a 072.2 072.a |
            grep ^t | awk -F'\t' '$2 ~ "NA"' | awk -F'\t' '$3 ~ "bicssc"'

    would get you almost there. TODO: Learn more awk.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'codes': NEPSubjectFilterCodes(date=self.date),
                'snaphost': NEPSnapshot(date=self.date)}

    @timed
    def run(self):
        with self.input().get('codes').open() as handle:
            filter_codes = map(string.strip, handle.readlines())

        with self.input().get('snaphost').open() as handle:
            with self.output().open('w') as output:
                reader = pymarc.MARCReader(handle, to_unicode=True)
                writer = pymarc.MARCWriter(output)
                for record in reader:
                    record = marcx.FatRecord.from_record(record)

                    # 240.a Uniform title
                    if record.has('240.a'):
                        continue

                    # 007 Physical Description Fixed Field, Text
                    if not record.test('007', lambda s: s.startswith('t')):
                        continue

                    # 072.2 Subject Category Code, Source
                    if record.test('072.2', lambda s: s.startswith('bicssc')):
                        for code in record.itervalues('072.a'):
                            if code in filter_codes:
                                writer.write(record.to_record())
                                break

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class NEPJson(NEPTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return NEPForUBL(date=self.date)

    @timed
    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                          input=self.input().path, date=self.closest())
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))

class NEPIndex(NEPTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'nep'
    doc_type = 'title'
    purge_existing_index = True

    settings = {
        "settings": {
            "number_of_shards": 5,
            "analysis": {
                "filter": {
                    "autocomplete_filter": {
                        # normal ngram vs edge?
                        "type":     "edge_ngram",
                        "min_gram": 1,
                        "max_gram": 20
                    }
                },
                "analyzer": {
                    "autocomplete": {
                        "type":      "custom",
                        "tokenizer": "standard",
                        "filter": [
                            # add more?
                            "lowercase",
                            "autocomplete_filter"
                        ]
                    }
                }
            }
        }
    }

    mapping = {
        'title': {
            'date_detection': False,
            '_id': {
                'path': 'content.001'
            },
            '_all': {
                'enabled': True,
                'term_vector': 'with_positions_offsets',
                'store': True
            },
            'properties': {
                'content': {
                    'properties': {
                        '245': {
                            'properties': {
                                # multifield?
                                'a': {
                                    'type': 'string',
                                    'index_analyzer': 'autocomplete',
                                    'search_analyzer': 'standard'
                                }
                            }
                        },
                        '100': {
                            'properties': {
                                'a': {
                                    'type': 'string',
                                    'index_analyzer': 'autocomplete',
                                    'search_analyzer': 'standard'
                                }
                            }
                        },
                        '700': {
                            'properties': {
                                'a': {
                                    'type': 'string',
                                    'index_analyzer': 'autocomplete',
                                    'search_analyzer': 'standard'
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return NEPJson(date=self.date)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Gallica.
"""

from __future__ import print_function
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.task import DefaultTask
from siskin.utils import iterfiles, date_range
import BeautifulSoup
import codecs
import collections
import datetime
import itertools
import json
import luigi
import marcx
import re
import tempfile

class BNFTask(DefaultTask):
    TAG = '020'

    def closest(self):
        return monthly(date=self.date)

class BNFHarvest(BNFTask):
    """
    Harvest BNF. Gallica.
    """
    def requires(self):
        return Executable(name='oaimi', message='https://github.com/miku/oaimi')

    def run(self):
        output = shellout("oaimi -root collection -verbose -set gallica:typedoc:partitions http://oai.bnf.fr/oai2/OAIHandler > {output}")
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

# class BNFHarvestChunk(BNFTask):
#     """ Harvest all files in chunks. """
#     begin = luigi.DateParameter(default=datetime.date(2011, 1, 1))
#     end = ClosestDateParameter(default=datetime.date.today())
#     prefix = luigi.Parameter(default="oai_dc")
#     collection = luigi.Parameter(default="gallica:typedoc:partitions")

#     @timed
#     def run(self):
#         """ Harvest files for a certain timeframe in a temporary
#         directory, then combine all records into a single file. """
#         stopover = tempfile.mkdtemp(prefix='tasktree-')
#         oai_harvest(url="http://oai.bnf.fr/oai2/OAIHandler",
#                     begin=self.begin, end=self.end, prefix=self.prefix,
#                     directory=stopover, collection=self.collection)
#         with self.output().open('w') as output:
#             output.write("""<collection
#                 xmlns="http://www.openarchives.org/OAI/2.0/"
#                 xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">""")
#             for path in iterfiles(stopover):
#                 with open(path) as handle:
#                     soup = BeautifulSoup.BeautifulStoneSoup(handle.read())
#                     for record in soup.findAll('record'):
#                         output.write(str(record)) # or unicode?
#             output.write('</collection>\n')

#     def output(self):
#         return luigi.LocalTarget(path=self.path(ext='xml', digest=True))

# class BNFJsonChunk(BNFTask):
#     """ Combine the chunks for a date range into a single file.
#     The filename will carry a name, that will only include year
#     and month for begin and end."""
#     begin = luigi.DateParameter(default=datetime.date(2011, 1, 1))
#     end = ClosestDateParameter(default=datetime.date.today())
#     prefix = luigi.Parameter(default="oai_dc")
#     collection = luigi.Parameter(default="gallica:typedoc:partitions")

#     def requires(self):
#         return BNFHarvestChunk(begin=self.begin, end=self.end,
#                                prefix=self.prefix, collection=self.collection)

#     @timed
#     def run(self):
#         """ TODO: this is much much too long. """
#         fr_en_map = {
#             u'domaine public': u'public domain',
#             u'musique manuscrite': u'manuscript music',
#             u"Notice d'ensemble : ": '',
#             u'Notice du catalogue : ': '',
#             u'Notice de recueil : ': '',
#         }

#         irregulars = {
#             u'Cellarius, Henri (neveu18..-18..? ; chorégraphe). Chorégraphe':
#                 u'Cellarius, Henri (18..-18..? ; chorégraphe). Chorégraphe',
#             u'Sophocle (0496?-0406 av. J.-C.). Auteur adapté':
#                 u'Sophocle (-0496?-0406). Auteur adapté',
#             u'Horace (0065-0008 av. J.-C.). Auteur du texte':
#                 u'Horace (-0065-0008). Auteur du texte',
#             u'Lacour (17..-ap. 1826). Librettiste':
#                 u'Lacour (17..-1826). Librettiste',
#             u'Eschyle (0525?-0456 av. J.-C.). Auteur adapté':
#                 u'Eschyle (-0525?-0456). Auteur adapté',
#             u'Benaut, Josse-François-Joseph (1743-1794). Compositeur. 1100':
#                 u'Benaut, Josse-François-Joseph (1743-1794). Compositeur',
#             u'Capron, Nicolas (1740?-1784). Compositeur. 1100':
#                 u'Capron, Nicolas (1740?-1784). Compositeur'
#         }

#         def parse_people(s):
#             if s in irregulars.keys():
#                 for match, replacement in irregulars.iteritems():
#                     s = s.replace(match, replacement)
#             if s.endswith(')'):
#                 if re.search(r'[0-9]\)$', s):
#                     s += '. TYPE'
#             if s in (u'Agence de publicité', u'Dédicataire', u'Parolier',
#                      u'Compositeur'):
#                 s = u'Anonyme. ' + s

#             # person name: no digits
#             a = r'(?P<a>[^0-9]+)'
#             # (year-year ; foo)
#             # d = r'\((?P<d>[0-9\.\-\? ]+)(?: ; [^0-9]+)?\)'
#             # (foo ; year-year ; bar)
#             # d = r'\(([^0-9]+ ; )?(?P<d>[0-9\.\-\? ]+)( ; [^0-9]+)?\)'
#             # (foo ; year-year ; bar)
#             d = r'\(([^0-9]+ ; )?(?P<d>[0-9\.\-\? ]+)( ; .*?)?\)'
#             # type of work: no digits
#             e = r'(?P<e>[^0-9]+)'
#             # type of work: no digits
#             # e = r'(?P<e>[^0-9]{5,})'

#             # name (years). type
#             test = re.match(r'%s( %s)?(\. %s)$' % (a, d, e), s)

#             if test:
#                 # io.append({ 'l':line, 'm':test.groupdict() })
#                 # k['t'].append(test.groupdict())
#                 sf = test.groupdict()
#             else:
#                 # io.append({ 'l':line })
#                 # k['f'].append(line)
#                 # no match: assume person name alone
#                 sf = {'a': s, 'd': '', 'e' : ''}
#                 # no match: assume person name alone
#                 # sf = {'a': line, 'd': '', 'e':''}
#             return sf

#         docs = []

#         with self.input().open() as handle:
#             soup = BeautifulSoup.BeautifulStoneSoup(handle.read())

#         for record in soup.findAll('record'):
#             doc = collections.defaultdict(list)

#             header = record.find('header')

#             # identifier
#             doc['oai'] = header.find('identifier').text
#             doc['id'] = doc['oai'].split('/')[-1] # too short?

#             if header.get('status') == 'deleted':
#                 self.logger.debug("Skipping deleted: {0}".format(doc['oai']))
#                 continue

#             metadata = record.find('metadata')

#             # title
#             for title in metadata.findAll('dc:title'):
#                 doc['title'].append(title.text)

#             # creator
#             for creator in metadata.findAll('dc:creator'):
#                 doc['creator'].append(parse_people(creator.text))

#             # subject
#             for subject in metadata.findAll('dc:subject'):
#                 doc['subject'].append(subject.text)

#             # publisher
#             for publisher in metadata.findAll('dc:publisher'):
#                 doc['publisher'].append(publisher.text)

#             # pubdate
#             for pubdate in metadata.findAll('dc:date'):
#                 doc['pubdate'].append(pubdate.text)

#             # description
#             for description in metadata.findAll('dc:description'):
#                 doc['description'].append(description.text)

#             # format
#             for format in metadata.findAll('dc:format'):
#                 if format.text.startswith('application/'):
#                     doc['mimetype'].append(format.text)
#                 else:
#                     doc['format'].append(format.text)

#             # relation
#             relations = set()
#             for rel in metadata.findAll('dc:relation'):
#                 translation = fr_en_map.get(rel.text, rel.text)
#                 if translation:
#                     relations.add(translation)
#             doc['relation'] = list(relations)

#             # score and pubtype
#             scores, pubtype = set(), set()
#             for dctype in metadata.findAll('dc:type'):
#                 translated = fr_en_map.get(dctype.text, '')
#                 if translated == 'manuscript music':
#                     scores.add(translated)
#                 else:
#                     pubtype.add(translated)

#             doc['score'] = list(scores)
#             doc['pubtype'] = list(pubtype)

#             # dc identifier
#             for dcid in metadata.findAll('dc:identifier'):
#                 if dcid.text.startswith('http://'):
#                     doc['url'].append(dcid.text)
#                 else:
#                     doc['identifier'].append(dcid.text)

#             for source in metadata.findAll('dc:source'):
#                 doc['source'].append(source.text)

#             rights = set()
#             for right in metadata.findAll('dc:rights'):
#                 rights.add(fr_en_map.get(right.text))

#             docs.append(doc)

#         # dump as json, one document per line
#         with self.output().open('w') as output:
#             for doc in docs:
#                 output.write(json.dumps(doc))
#                 output.write('\n')

#     def output(self):
#         return luigi.LocalTarget(path=self.path(ext='json', digest=True))

# class BNFMarcChunk(BNFTask):
#     """ Convert a JSON snippet into MARC. """
#     begin = luigi.DateParameter(default=datetime.date(1970, 1, 1))
#     end = ClosestDateParameter(default=datetime.date.today())
#     prefix = luigi.Parameter(default="oai_dc")
#     collection = luigi.Parameter(default="gallica:typedoc:partitions")

#     def requires(self):
#         return BNFJsonChunk(begin=self.begin, end=self.end,
#                             prefix=self.prefix, collection=self.collection)

#     @timed
#     def run(self):
#         _, stopover = tempfile.mkstemp(prefix="tasktree-")
#         with self.input().open() as handle:
#             with codecs.open(stopover, 'w', 'utf-8') as output:
#                 for line in handle:
#                     record = marcx.FatRecord(to_unicode=True, force_utf8=True)
#                     doc = json.loads(line.encode('utf-8'))

#                     record.add('001', data=doc['id'])
#                     record.add('037', n=doc['oai'])
#                     record.add('042', a='dc')

#                     people = itertools.chain(doc.get('creator', []),
#                                              doc.get('contributors', []))
#                     for i, person in enumerate(people):
#                         a = person.get('a') or ''
#                         d = person.get('d') or ''
#                         e = person.get('e') or ''
#                         if i == 0:
#                             # only add the first creator as 100
#                             record.add('100', a=a, d=d, e=e)
#                         else:
#                             # all others need to go to 700
#                             record.add('700', a=a, d=d, e=e)

#                     for i, title in enumerate(doc.get('title', [])):
#                         if i == 0:
#                             record.add('245', a=title)
#                         else:
#                             record.add('242', a=title)

#                     for publisher in doc.get('publisher', []):
#                         record.add('260', b=publisher)

#                     for pubdate in doc.get('pubdate', []):
#                         record.add('260', c=pubdate)

#                     for description in doc.get('description', []):
#                         record.add('520', a=description)

#                     for score in doc.get('score', []):
#                         record.add('254', a=score)

#                     for pubtype in doc.get('pubtype', []):
#                         record.add('655', indicators='7 ', a=pubtype)

#                     if 'url' in doc:
#                         subs = collections.defaultdict(list)
#                         if 'mimetype' in doc:
#                             subs['q'] = doc.get('mimetype')[0]
#                         if 'rights' in doc:
#                             subs['z'] = doc.get('rights')[0]
#                         subs['u'] = doc.get('url')[0]
#                         record.add('856', indicators='40', **subs)

#                     for subject in doc.get('subject', []):
#                         record.add('609', a=subject)

#                     for ident in doc.get('identifier', []):
#                         record.add('609', a=ident)

#                     if len(doc.get('relation', [])) > 0:
#                         record.add('786', i=[r for r in doc.get('relation', [])])

#                     output.write(record.as_marc().decode('utf-8'))

#         luigi.File(stopover).move(self.output().path)

#     def output(self):
#         return luigi.LocalTarget(path=self.path(ext='mrc', digest=True))

# class BNFMarcHarvest(luigi.WrapperTask, BNFTask):
#     """ Harvest BNF, cut all dates to the first of month. Example:
#     2013-05-12 .. 2013-09-11 would be translated to
#     2013-05-01 .. 2013-09-01"""
#     begin = luigi.DateParameter(default=datetime.date(2011, 1, 1))
#     end = ClosestDateParameter(default=datetime.date.today())
#     prefix = luigi.Parameter(default="oai_dc")
#     collection = luigi.Parameter(default="gallica:typedoc:partitions")

#     def requires(self):
#         """ Only require up to the last full month. """
#         begin = datetime.date(self.begin.year, self.begin.month, 1)
#         end = datetime.date(self.end.year, self.end.month, 1)
#         if end < self.begin:
#             raise RuntimeError('Invalid range: %s - %s' % (begin, end))
#         dates = date_range(begin, end, 1, 'months')
#         for i, _ in enumerate(dates[:-1]):
#             yield BNFMarcChunk(begin=dates[i], end=dates[i + 1],
#                                prefix=self.prefix, collection=self.collection)

#     def output(self):
#         return self.input()

# class BNFCombine(BNFTask):
#     """ Combine all MARC files into one. """
#     begin = luigi.DateParameter(default=datetime.date(2011, 1, 1))
#     date = ClosestDateParameter(default=datetime.date.today())
#     prefix = luigi.Parameter(default="oai_dc")
#     collection = luigi.Parameter(default="gallica:typedoc:partitions")

#     def requires(self):
#         return BNFMarcHarvest(begin=self.begin, end=self.date,
#                               prefix=self.prefix, collection=self.collection)

#     @timed
#     def run(self):
#         _, combined = tempfile.mkstemp(prefix='tasktree')
#         for target in self.input():
#             shellout("cat {input} >> {output}", input=target.path,
#                      output=combined)
#         output = shellout("marcuniq -o {output} {input}", input=combined)
#         luigi.File(output).move(self.output().path)

#     def output(self):
#         return luigi.LocalTarget(path=self.path(ext='mrc', digest=True))

# class BNFJson(BNFTask):
#     """ Convert to JSON. """
#     begin = luigi.DateParameter(default=datetime.date(2011, 1, 1))
#     date = ClosestDateParameter(default=datetime.date.today())
#     prefix = luigi.Parameter(default="oai_dc")
#     collection = luigi.Parameter(default="gallica:typedoc:partitions")

#     def requires(self):
#         return BNFCombine(begin=self.begin, date=self.date, prefix=self.prefix,
#                           collection=self.collection)

#     def run(self):
#         tmp = shellout("marctojson -m date={date} {input} > {output}",
#                        input=self.input().path, date=self.date)
#         luigi.File(tmp).move(self.output().path)

#     def output(self):
#         """ Use monthly files. """
#         return luigi.LocalTarget(path=self.path(ext='ldj', digest=True))

# class BNFIndex(BNFTask, CopyToIndex):
#     begin = luigi.DateParameter(default=datetime.date(2011, 1, 1))
#     date = ClosestDateParameter(default=datetime.date.today())
#     prefix = luigi.Parameter(default="oai_dc")
#     collection = luigi.Parameter(default="gallica:typedoc:partitions")

#     index = 'bnf'
#     doc_type = 'title'
#     purge_existing_index = True

#     mapping = {'title': {'date_detection': False,
#                           '_id': {'path': 'content.001'},
#                           '_all': {'enabled': True,
#                                    'term_vector': 'with_positions_offsets',
#                                    'store': True}}}

#     def update_id(self):
#         """ This id will be a unique identifier for this indexing task."""
#         return self.effective_task_id()

#     def requires(self):
#         return BNFJson(begin=self.begin, date=self.date, prefix=self.prefix,
#                        collection=self.collection)



# coding: utf-8

"""
Web of Science.
"""

from gluish.utils import shellout
from siskin.task import DefaultTask
from siskin.sources.crossref import CrossrefDOIAndISSNList
from siskin.sources.doaj import DOAJDOIList, DOAJISSNList
from gluish.format import TSV
import datetime
import luigi
import re

class WOSTask(DefaultTask):
    TAG = 'wos'

    def closest(self):
        return monthly(date=self.date)

class WOSPublicationList(WOSTask):
    """
    Publication list. PDF -> txt.
    """
    date = luigi.Parameter(default=datetime.date.today())

    def run(self):
        output = shellout("wget -O {output} http://ip-science.thomsonreuters.com/mjl/publist_sciex.pdf")
        output = shellout("pdftotext -layout {input} {output}", input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

class WOSISSNList(WOSTask):
    """
    Rough ISSN list.
    """
    date = luigi.Parameter(default=datetime.date.today())

    def requires(self):
        return WOSPublicationList(date=self.date)

    def run(self):
        issns = set()
        with self.input().open() as handle:
            for line in handle:
                for issn in re.findall(r'[0-9]{4}-[0-9]{3}[0-9X]', line):
                    issns.add(issn)

        with self.output().open('w') as output:
            for issn in sorted(issns):
                output.write_tsv(issn)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WOSCrossrefMatch(WOSTask):
    """
    Find all ISSNs for Web of Science export.

    The input --file must contain one DOI per line:

    005410.001002/14651858.CD14005465.pub14651852
    10.1002/ange.201106065
    10.1007/BFb0061839.56
    ..
    """
    date = luigi.DateParameter(default=datetime.date.today())
    file = luigi.Parameter(description='web of science doi list', significant=False)

    def requires(self):
        return CrossrefDOIAndISSNList(date=self.date)

    def run(self):
        # wos contains all WOS DOIs, found keeps the DOIs, that could be found in Crossref
        wos, found = set(), set()

        with open(self.file) as handle:
            for line in handle:
                wos.add(line.strip().lower())

        with self.input().open() as handle:
            with self.output().open('w') as output:
                for line in handle:
                    parts = [s.strip('"') for s in line.strip().split(',')]
                    doi, issns = parts[0].lower(), parts[1:]
                    if doi in wos:
                        output.write_tsv('OK', doi, *issns)
                        found.add(doi)

                for id in wos.difference(found):
                    output.write_tsv('NOT_FOUND', id)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WOSMatchDOAJAndReferenceDOI(WOSTask):
    """
    For all references in WOS export, check if DOI is in DOAJ.
    """
    date = luigi.DateParameter(default=datetime.date.today())
    file = luigi.Parameter(description='web of science doi list (references)', significant=False)

    def requires(self):
        return DOAJDOIList(date=self.date)

    def run(self):
        reference_dois = set()
        with open(self.file) as handle:
            for line in handle:
                reference_dois.add(line.strip())

        with self.input().open() as handle:
            with self.output().open('w') as output:
                for row in handle:
                    parts = row.split('\t')
                    if len(parts) < 0:
                        continue
                    doi = parts[0].strip()
                    if doi in reference_dois:
                        output.write_tsv('OPEN_ACCESS', doi)
                    else:
                        output.write_tsv('NOT_IN_DOAJ', doi)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WOSISSNForDOI(WOSTask):
    """
    Extract ISSN for a list of DOIs.
    """
    date = luigi.DateParameter(default=datetime.date.today())
    file = luigi.Parameter(description='web of science doi list (references)', significant=False)

    def requires(self):
        return CrossrefDOIAndISSNList(date=self.date)

    def run(self):
        reference_dois = set()
        with open(self.file) as handle:
            for line in handle:
                reference_dois.add(line.strip())

        with self.input().open() as handle:
            with self.output().open('w') as output:
                for i, line in enumerate(handle):
                    if i % 1000000 == 0:
                        self.logger.info('checked %s dois' % i)
                    line = line.replace('","', '\t').replace('"', '').strip()
                    parts = line.split('\t')
                    if len(parts) < 2:
                        continue
                    doi, issns = parts[0], parts[1:]
                    if doi in reference_dois:
                        output.write_tsv(doi, *issns)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WOSISSNInDOAJ(WOSTask):
    """
    For all references in WOS export, check if DOI is in DOAJ.
    """
    date = luigi.DateParameter(default=datetime.date.today())
    file = luigi.Parameter(description='web of science doi list (references)', significant=False)

    def requires(self):
        return {
            'wos-doi-list': WOSISSNForDOI(file=self.file, date=self.date),
            'doaj-issns': DOAJISSNList(date=self.date),
        }

    def run(self):
        doaj_issns = set()
        with self.input().get('doaj-issns').open() as handle:
            for line in handle:
                doaj_issns.add(line.strip())

        with self.input().get('wos-doi-list').open() as handle:
            with self.output().open('w') as output:
                for line in handle:
                    fields = line.strip().split('\t')
                    doi = fields[0]
                    for issn in fields[1:]:
                        if issn in doaj_issns:
                            output.write_tsv('IN_DOAJ', doi, issn)
                        else:
                            output.write_tsv('NOT_IN_DOAJ', doi, issn)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301,C0103,W0614,W0401
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Various utilities.
"""

from __future__ import print_function
from dateutil import relativedelta
from luigi.task import Register, flatten
from siskin import __version__
import collections
import cStringIO as StringIO
import errno
import functools
import hashlib
import itertools
import json
import luigi
import os
import pprint
import random
import requests
import string
import sys
import tempfile

MAN_HEADER = r"""

  \V/    \V/    \V/    \V/    \V/    \V/    \V/    \V/    \V/    \V/    \V/
---o------o------o------o------o------o------o------o------o------o------o-----

+ Siskin
"""

class SetEncoder(json.JSONEncoder):
    """ Helper to encode python sets into JSON lists. """
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)

def wc(path):
    """
    Like `wc -l`.
    """
    with open(path) as handle:
        num_lines = sum(1 for line in handle)
    return num_lines

def date_range(start_date, end_date, increment, period):
    """
    Generate `date` objects between `start_date` and `end_date` in `increment`
    `period` intervals.
    """
    result = []
    nxt = start_date
    delta = relativedelta.relativedelta(**{period:increment})
    while nxt <= end_date:
        result.append(nxt)
        nxt += delta
    return result

def copyregions(src, dst, seekmap):
    """
    Copy regions of a source file to a target. The regions are given in the
    iterable `seekmap`, which must contain (offset, length) tuples.

    `src` (r) and `dst` (w) must be open files handles.

    More: http://stackoverflow.com/q/18551592/89391
    """
    for offset, length in sorted(seekmap):
        src.seek(offset)
        dst.write(src.read(length))

class memoize(object):
    '''Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).
    '''
    def __init__(self, func):
        self.func = func
        self.cache = {}
    def __call__(self, *args):
        if not isinstance(args, collections.Hashable):
            # uncacheable. a list, for instance.
            # better to not cache than blow up.
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value
    def __repr__(self):
        '''Return the function's docstring.'''
        return self.func.__doc__
    def __get__(self, obj, objtype):
        '''Support instance methods.'''
        return functools.partial(self.__call__, obj)

def iterfiles(directory='.', fun=None):
    """
    Yield paths below a directory, optionally filter the path through a function
    given in `fun`.
    """
    if fun is None:
        fun = lambda path: True
    for root, dirs, files in os.walk(directory):
        for f in files:
            path = os.path.join(root, f)
            if fun(path):
                yield path

def generate_tasks_manual():
    """ Return a formatted listing of all tasks with their descriptions. """
    from siskin.sources import *
    from siskin.workflows import *

    output = StringIO.StringIO()
    # task_tuples = sorted(Register.get_reg().iteritems())
    task_names = Register.task_names()
    output.write(MAN_HEADER)
    output.write('  {0} tasks found\n\n'.format(len(task_names)))

    for name in task_names:
        klass = Register.get_task_cls(name)
        doc = klass.__doc__ or "@TODO: docs"
        output.write('{0} {1}\n'.format(name, doc))

        try:
            deps = flatten(klass().requires())
        except Exception:
            # TODO: tasks that have required arguments will fail here
            formatted = "\tUnavailable since task has required parameters."
        else:
            formatted = '\t{0}'.format(pprint.pformat(deps).replace('\n', '\n\t'))
        output.write('\n\tDependencies ({0}):\n\n{1}\n\n'.format(len(deps), formatted))

    return output.getvalue()

def random_string(length=16):
    """
    Return a random string (upper and lowercase letters) of length `length`,
    defaults to 16.
    """
    return ''.join(random.choice(string.letters) for _ in range(length))

def pairwise(obj):
    """ Iterator over a iterable in steps of two. """
    iterable = iter(obj)
    return itertools.izip(iterable, iterable)


def nwise(iterable, n=2):
    """
    Generalized :func:`pairwise`.
    Split an iterable after every `n` items.
    """
    i = iter(iterable)
    piece = tuple(itertools.islice(i, n))
    while piece:
        yield piece
        piece = tuple(itertools.islice(i, n))

def get_task_import_cache():
    """
    Load `taskname: modulename` mappings from dictionary. Return a tuple containing
    the dictionary and the path to the cache file.
    """
    task_import_cache = None
    path = os.path.join(tempfile.gettempdir(), 'siskin_task_import_cache_%s' % __version__)
    if not os.path.exists(path):
        from siskin.sources import *
        from siskin.workflows import *
        with open(path, 'w') as output:
            task_import_cache = dict([(name, Register.get_task_cls(name).__module__) for name in Register.task_names() if name[0].isupper()])
            json.dump(task_import_cache, output)

    if task_import_cache is None:
        with open(path) as handle:
            try:
                task_import_cache = json.load(handle)
            except Exception as err:
                print("failed load task import cache, try removing %s and then try again" % path, file=sys.stderr)
                sys.exit(1)

    return task_import_cache, path

class URLCache(object):
    """ A simple URL *content* cache. Stores everything on the filesystem.
    Content is first written to a temporary file and then renamed.
    With concurrent requests for the same URL, the last one wins.
    Raises exception on any statuscode >= 400. """

    def __init__(self, directory=None):
        self.directory = directory or tempfile.gettempdir()
        self.sess = requests.session()

    def get_cache_file(self, url):
        digest = hashlib.sha1(url).hexdigest()
        d0, d1, d2 = digest[:2], digest[2:4], digest[4:6]
        path = os.path.join(self.directory, d0, d1, d2)
        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except OSError as e:
                if e.errno == errno.EEXIST:
                    pass
                else:
                    raise
        return os.path.join(path, digest)

    def is_cached(self, url):
        return os.path.exists(self.get_cache_file(url))

    def get(self, url):
        if not self.is_cached(url):
            r = self.sess.get(url)
            if r.status_code >= 400:
                raise RuntimeError('%s on %s' % (r.status_code, url))
            with tempfile.NamedTemporaryFile(delete=False) as output:
                output.write(r.text.encode('utf-8'))
            os.rename(output.name, self.get_cache_file(url))

        with open(self.get_cache_file(url)) as handle:
            return handle.read()

class ElasticsearchMixin(luigi.Task):
    """ A small mixin for tasks that require an ES connection. """
    es_host = luigi.Parameter(default='localhost', significant=False,
                              description='elasticsearch host')
    es_port = luigi.IntParameter(default=9200, significant=False,
                                 description='elasticsearch port')



#!/usr/bin/env python
# coding: utf-8
#
#  Copyright 2016 by Leipzig University Library, http://ub.uni-leipzig.de
#                    The Finc Authors, http://finc.info
#                    Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Journaltocs API. http://www.journaltocs.ac.uk

26,120 journal TOCs (2,728 publishers).

[journaltocs]

user = theuserid
"""

from gluish.utils import shellout
from siskin.configuration import Config
from siskin.task import DefaultTask
import luigi

config = Config.instance()

class JournaltocsTask(DefaultTask):
    """
    Base task.
    """
    TAG = 'journaltocs'

class JournaltocsJournalHarvest(JournaltocsTask):
    """
    A sample harvest of the API.
    """
    query = luigi.Parameter(description='query string')

    def run(self):
        output = shellout(""" curl --fail "http://www.journaltocs.ac.uk/api/journals/{query}?user={user}" > {output}""", query=self.query, user=config.get('journaltocs', 'user'))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True, ext='xml'))

class JournaltocsHarvest(JournaltocsTask, luigi.WrapperTask):
    """
    Harvest a bunch of keywords.
    """

    def requires(self):
        queries = ["*ancient", "*modern*", "*science*", "*engineering*", "*philosophy*", "*medicine*", "*sociology*", "*language*",
                   "*german*", "*english*", "*american*", "*french*", "*library*", "*europe*", "*asia*", "*afric*"]
        for query in queries:
            yield JournaltocsJournalHarvest(query=query)

    def output(self):
        return self.input()



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

from __future__ import print_function
from gluish.common import Executable
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.task import DefaultTask
import datetime
import luigi
import urllib

class RISMTask(DefaultTask):
    TAG = '014'

    def closest(self):
        return datetime.date(2014, 1, 1)

class RISMSync(RISMTask):
    """ Get at most n records (defaults to 20000). """
    date = ClosestDateParameter(default=datetime.date.today())
    maximum = luigi.IntParameter(default=20000, significant=False)

    def requires(self):
        return Executable(name='wget')

    @timed
    def run(self):
        base = 'http://service1.rism.info/opac'
        params = {
            'operation': 'searchRetrieve',
            'version': '1.1',
            'query': 'source=http*',
            'maximumRecords': self.maximum,
        }
        output = shellout("""wget --retry-connrefused -O {output} "{base}?{params}" """,
                          base=base, params=urllib.urlencode(params))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))

class RISMMarc(RISMTask):
    """ Convert IMSLP XML (trans, antizebra) to MARCXML, then to MARC """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'file': RISMSync(date=self.date),
                'apps': [Executable(name='xsltproc'),
                         Executable(name='yaz-marcdump'),
                         Executable(name='marcuniq')]}

    @timed
    def run(self):
        # first trans.xsl
        output = shellout("xsltproc {stylesheet} {input} > {output}",
                          input=self.input().get('file').path,
                          stylesheet=self.assets('trans.xsl'))
        # then antizebra.xsl
        output = shellout("xsltproc {stylesheet} {input} > {output}",
                          input=output, stylesheet=self.assets('antizebra.xsl'))
        # then binary marc
        output = shellout("yaz-marcdump -i marcxml -o marc {input} > {output}",
                          input=output)
        # then marcuniq
        output = shellout("marcuniq -o {output} {input}", input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class RISMJson(RISMTask):
    """ Convert MARC to JSON. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return RISMMarc(date=self.date)

    @timed
    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                          input=self.input().path, date=self.closest())
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class RISMIndex(RISMTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'rism'
    doc_type = 'title'

    mapping = {'title': {'date_detection': False,
                          '_id': {'path': 'content.001'},
                          '_all': {'enabled': True,
                                   'term_vector': 'with_positions_offsets',
                                   'store': True}}}

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return RISMJson(date=self.date)



# coding: utf-8

"""
Support for importing all workflows at once.
"""

import os
import glob

modules = glob.glob("%s/*.py" % (os.path.dirname(__file__)))
__all__ = [os.path.basename(f)[:-3] for f in modules]



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2016 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Elsevier jounrals. Refs. #6975.

Configuration keys:

[elsevierjournals]

ftp-host = sftp://host.name
ftp-username = username
ftp-password = password
ftp-path = /
ftp-pattern = *

"""

from BeautifulSoup import BeautifulStoneSoup
from gluish.common import Executable
from gluish.format import TSV
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.benchmark import timed
from siskin.common import FTPMirror
from siskin.configuration import Config
from siskin.task import DefaultTask
from siskin.utils import iterfiles
import base64
import collections
import datetime
import json
import luigi
import os
import re

config = Config.instance()

class ElsevierJournalsTask(DefaultTask):
    """ Jstor base. """
    TAG = '085'

    def closest(self):
        return weekly(self.date)

class ElsevierJournalsPaths(ElsevierJournalsTask):
    """
    Sync.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    max_retries = luigi.IntParameter(default=10, significant=False)
    timeout = luigi.IntParameter(default=20, significant=False, description='timeout in seconds')

    def requires(self):
        return FTPMirror(host=config.get('elsevierjournals', 'ftp-host'),
                         username=config.get('elsevierjournals', 'ftp-username'),
                         password=config.get('elsevierjournals', 'ftp-password'),
                         pattern=config.get('elsevierjournals', 'ftp-pattern'),
                         max_retries=self.max_retries,
                         timeout=self.timeout)

    @timed
    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class ElsevierJournalsExpand(ElsevierJournalsTask):
    """
    Expand all tar files.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    tag = luigi.Parameter(default='SAXC0000000000002')

    def requires(self):
        return ElsevierJournalsPaths(date=self.date)

    @timed
    def run(self):
        shellout("mkdir -p {dir}", dir=self.taskdir())
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                if not self.tag in row.path:
                    continue
                if row.path.endswith('tar'):
                    shellout("tar -xvf {tarfile} -C {dir}", tarfile=row.path, dir=self.taskdir())

        with self.output().open('w') as output:
            for dirName, subdirList, fileList in os.walk(self.taskdir()):
                for fname in fileList:
                    output.write_tsv(os.path.join(dirName, fname))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class ElsevierJournalsIntermediateSchema(ElsevierJournalsTask):
    """
    All in one task for now.

    Test indexing:

        $ span-solr $(taskoutput ElsevierJournalsIntermediateSchema) > x.ldj
        $ solrbulk x.ldj

    """
    date = ClosestDateParameter(default=datetime.date.today())
    tag = luigi.Parameter(default='SAXC0000000000002')

    def requires(self):
        return ElsevierJournalsExpand(date=self.date, tag=self.tag)

    @timed
    def run(self):
        pattern = re.compile('(?P<base>.*)/(?P<tag>%s)/(?P<issn>.*)/(?P<issue>.*)/(?P<document>.*)/main.xml' % self.tag)

        # doctree groups main files under issue files: {"issue.xml": ["main.xml", "main.xml", ....]}
        doctree = collections.defaultdict(list)

        # the dataset.xml contains the journal titles
        datasetpath = os.path.join(os.path.dirname(self.input().path), self.tag, 'dataset.xml')
        self.logger.info('dataset: %s' % datasetpath)

        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                if not pattern.match(row.path):
                    continue

                issuepath = "%s/issue.xml" % '/'.join(row.path.split('/')[:-2])
                if os.path.exists(issuepath):
                    doctree[issuepath].append(row.path)
                else:
                    self.logger.warning('issue.xml not in expected place for %s' % row.path)

        if not doctree:
            raise RuntimeError('no documents found')

        with open(datasetpath) as handle:
            dataset = BeautifulStoneSoup(handle.read(), convertEntities=BeautifulStoneSoup.HTML_ENTITIES)

        # map ISSN to journal title
        titlemap = collections.defaultdict(str)
        for props in dataset.findAll('journal-issue-properties'):
            try:
                titlemap[props.find('issn').text] = props.find('collection-title').text
            except AttributeError as err:
                self.logger.warning('cannot find issn or collection title: %s, %s, %s' % (props.find('issn'), props.find('collection-title'), err))

        with self.output().open('w') as output:
            for issuepath, docs in doctree.iteritems():

                with open(issuepath) as handle:
                    issue = BeautifulStoneSoup(handle.read(), convertEntities=BeautifulStoneSoup.HTML_ENTITIES)

                for docpath in docs:
                    with open(docpath) as fh:
                        doc = BeautifulStoneSoup(fh.read(), convertEntities=BeautifulStoneSoup.HTML_ENTITIES)

                    intermediate = {
                        'finc.format': 'ElectronicArticle',
                        'finc.mega_collection': 'Elsevier Journals',
                        'finc.source_id': '85',
                        'rft.genre': 'article',
                        'rft.issn': [node.text for node in issue.findAll('ce:issn')],
                        'doi': doc.find('ce:doi').text,
                        'url': ['http://dx.doi.org/%s' % doc.find('ce:doi').text],
                        'languages': ['eng'],
                        'ris.type': 'EJOUR',
                        'version': '0.9',
                    }

                    if doc.find('ce:title'):
                        intermediate['rft.atitle'] = doc.find('ce:title').text

                    intermediate['rft.jtitle'] = titlemap[intermediate['rft.issn'][0]]
                    intermediate['finc.record_id'] = base64.b64encode(intermediate['url'][0]).rstrip("=")

                    if doc.find('ce:abstract'):
                        abstract = doc.find('ce:abstract').getText()
                        if abstract.startswith('Abstract'):
                            abstract = abstract.replace('Abstract', '', 1)
                        if abstract.startswith(u'Highlights•'):
                            abstract = abstract.replace(u'Highlights•', '', 1)
                        if abstract.startswith(u'SummaryBackground'):
                            abstract = abstract.replace(u'SummaryBackground', '', 1)
                        abstract = abstract.replace(u'•', ' ')
                        intermediate['abstract'] = abstract

                    authors = []
                    for author in doc.findAll('ce:author'):
                        au = {}
                        given, surname = author.find('ce:given-name'), author.find('ce:surname')
                        if given:
                            au.update({'rft.aufirst': given.text})
                        if surname:
                            au.update({'rft.aulast': surname.text})
                        if surname and given:
                            au.update({'rft.au': surname.text + ", " + given.text})
                        authors.append(au)
                    intermediate['authors'] = authors

                    keywords = []
                    for kw in doc.findAll('ce:keywords'):
                        keywords.append(kw.find('ce:text').text)
                    intermediate['x.subjects'] = keywords

                    # page numbers
                    for item in issue.findAll('ce:include-item'):
                        doi = item.find('ce:doi').text
                        if doi == intermediate['doi']:
                            first, last = item.find('ce:first-page'), item.find('ce:last-page')
                            if first:
                                intermediate['rft.spage'] = ''.join(c for c in first.text if c.isdigit())
                            if last:
                                intermediate['rft.epage'] = ''.join(c for c in last.text if c.isdigit())
                            if first and last:
                                try:
                                    intermediate['rft.pages'] = str(int(intermediate['rft.epage']) - int(intermediate['rft.spage']))
                                    intermediate['rft.tpages'] = intermediate['rft.pages']
                                except ValueError as err:
                                    self.logger.warning('cannot parse page number %s: %s-%s' % (doi, first.text, last.text))

                    # volume, issue, date
                    if issue.find('vol-first'):
                        intermediate['rft.volume'] = issue.find('vol-first').text
                    if issue.find('iss-first'):
                        intermediate['rft.issue'] = issue.find('iss-first').text
                    if issue.find('start-date'):
                        date = issue.find('start-date').text
                        if len(date) == 4:
                            intermediate['rft.date'] = "%s-01-01" % (date)
                        elif len(date) == 6:
                            intermediate['rft.date'] = "%s-%s-01" % (date[:4], date[4:6])
                        else:
                            raise ValueError("unknown date format: %s" % date)

                    output.write(json.dumps(intermediate))
                    output.write("\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class ElsevierJournalsSolr(ElsevierJournalsTask):
    """
    Create something solr importable. Attach a single ISIL to all records.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    tag = luigi.Parameter(default='SAXC0000000000002')

    def requires(self):
        return ElsevierJournalsIntermediateSchema(date=self.date, tag=self.tag)

    @timed
    def run(self):
        output = shellout("""span-tag -c <(echo '{{"DE-15": {{"any": {{}}}}}}') {input} > {output}""", input=self.input().path)
        output = shellout("span-solr {input} > {output}", input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))



# coding: utf-8

"""
Common tasks.
"""

from gluish.common import Executable
from gluish.utils import shellout
from gluish.format import TSV
from siskin.task import DefaultTask
from siskin.utils import iterfiles, random_string
import hashlib
import luigi
import os
import pipes

class CommonTask(DefaultTask):
    """
    A base class for common classes. These artefacts will be written to the
    systems tempdir.
    """
    TAG = 'common'

class Directory(luigi.Task):
    """ Create directory or fail. """
    path = luigi.Parameter(description='directory to create')

    def run(self):
        try:
            os.makedirs(self.path)
        except OSError as err:
            if err.errno == 17:
                # file exists, this can happen in parallel execution evns
                pass
            else:
                raise RuntimeError(err)

    def output(self):
        return luigi.LocalTarget(self.path)

class FTPMirror(CommonTask):
    """
    A generic FTP directory sync. Required lftp (http://lftp.yar.ru/).
    The output of this task is a single file, that contains the paths
    to all the mirrored files.
    """
    host = luigi.Parameter()
    username = luigi.Parameter(default='anonymous')
    password = luigi.Parameter(default='')
    pattern = luigi.Parameter(default='*', description="e.g. '*leip_*.zip'")
    base = luigi.Parameter(default='.')
    indicator = luigi.Parameter(default=random_string())
    max_retries = luigi.IntParameter(default=5, significant=False)
    timeout = luigi.IntParameter(default=10, significant=False, description='timeout in seconds')
    exclude_glob = luigi.Parameter(default="", significant=False, description='globs to exclude')

    def requires(self):
        return Executable(name='lftp', message='http://lftp.yar.ru/')

    def run(self):
        """ The indicator is always recreated, while the subdir
        for a given (host, username, base, pattern) is just synced. """
        base = os.path.dirname(self.output().path)
        subdir = hashlib.sha1('{host}:{username}:{base}:{pattern}'.format(
            host=self.host, username=self.username, base=self.base,
            pattern=self.pattern)).hexdigest()
        # target is the root of the mirror
        target = os.path.join(base, subdir)
        if not os.path.exists(target):
            os.makedirs(target)

        exclude_glob = ""
        if not self.exclude_glob == "":
            exclude_glob = "--exclude-glob %s" % self.exclude_glob

        command = """lftp -u {username},{password}
        -e "set net:max-retries {max_retries}; set net:timeout {timeout}; mirror --verbose=0
        --only-newer {exclude_glob} -I {pattern} {base} {target}; exit" {host}"""

        shellout(command, host=self.host, username=pipes.quote(self.username),
                 password=pipes.quote(self.password),
                 pattern=pipes.quote(self.pattern),
                 target=pipes.quote(target),
                 base=pipes.quote(self.base),
                 max_retries=self.max_retries,
                 timeout=self.timeout,
                 exclude_glob=exclude_glob)

        with self.output().open('w') as output:
            for path in iterfiles(target):
                self.logger.debug("Mirrored: %s" % path)
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)


class FTPFile(CommonTask):
    """ Just require a single file from an FTP server. """
    host = luigi.Parameter()
    username = luigi.Parameter()
    password = luigi.Parameter()
    filepath = luigi.Parameter()

    def requires(self):
        return Executable(name='lftp')

    def run(self):
        command = """lftp -u {username},{password}
        -e "set net:max-retries 5; set net:timeout 10; get -c
        {filepath} -o {output}; exit" {host}"""

        output = shellout(command, host=self.host,
                          username=pipes.quote(self.username),
                          password=pipes.quote(self.password),
                          filepath=pipes.quote(self.filepath))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True, ext=None))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Yago.
"""

from siskin.benchmark import timed
from gluish.intervals import yearly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask
import datetime
import luigi

class YagoTask(DefaultTask):
    TAG = 'yago'

    def closest(self):
        return yearly(self.date)

class YagoDump(YagoTask):
    """ Download and extract Yago dump. Might 404 from time to time. """
    date = ClosestDateParameter(default=datetime.date.today())

    @timed
    def run(self):
        url = "http://www.mpi-inf.mpg.de/yago-naga/yago/download/yago/yago2s_ttl.7z"
        output = shellout("""wget --retry-connrefused -O {output} {url}""", url=url)
        output = shellout("""7z e -so {input} 2> /dev/null > {output}""", input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ttl'))



# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Electronic Resource Management System Based on Linked Data Technologies.

http://amsl.technology

Config:

[amsl]

isil-rel = https://x.com/static/about.json
holdings = https://x.com/inhouseservices/list?do=holdings
uri-download-prefix = https://x.y.z/OntoWiki/files/get?setResource=

"""

from gluish.utils import shellout
from siskin.configuration import Config
from siskin.task import DefaultTask
from siskin.utils import SetEncoder
import collections
import datetime
import json
import luigi
import tempfile
import zipfile

config = Config.instance()

class AMSLTask(DefaultTask):
    TAG = 'amsl'

class AMSLCollections(AMSLTask):
    """ Get a list of ISIL, source and collection choices via JSON API. """

    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("""curl --fail "{link}" > {output} """, link=config.get('amsl', 'isil-rel'))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

class AMSLCollectionsISIL(AMSLTask):
    """
    Per ISIL list of collections.
    """
    date = luigi.DateParameter(default=datetime.date.today())
    isil = luigi.Parameter(description='ISIL, case sensitive')
    shard = luigi.Parameter(default='UBL-ai', description='only collect items for this shard')

    def requires(self):
        return AMSLCollections(date=self.date)

    def run(self):
        with self.input().open() as handle:
            c = json.load(handle)
        scmap = collections.defaultdict(set)
        for item in c:
            if not item['shardLabel_str'] == self.shard:
                continue
            if not item['isil_str'] == self.isil:
                continue
            scmap[item['sourceID']].add(item['collectionLabel'].strip())
        if not scmap:
            raise RuntimeError('no collections found for ISIL: %s' % self.isil)

        with self.output().open('w') as output:
            output.write(json.dumps(scmap, cls=SetEncoder) + "\n")

    def output(self):
        return luigi.LocalTarget(path=self.path())

class AMSLHoldings(AMSLTask):
    """
    Download AMSL tasks.
    """
    date = luigi.Parameter(default=datetime.date.today())

    def run(self):
        output = shellout(""" curl --fail "{link}" > {output} """, link=config.get('amsl', 'holdings'))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

class AMSLHoldingsISILList(AMSLTask):
    """
    Return a list of ISILs that are returned by the API.
    """
    date = luigi.Parameter(default=datetime.date.today())

    def requires(self):
        return AMSLHoldings(date=self.date)

    def run(self):
        output = shellout("jq -r '.[].ISIL' {input} | sort > {output}", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

class AMSLHoldingsFile(AMSLTask):
    """
    Access AMSL files/get?setResource= facilities.

    The output is probably zipped.

    One ISIL can have multiple files.

    If the file is Zipped, it will be decompressed.
    """
    isil = luigi.Parameter(description='ISIL, case sensitive')
    date = luigi.Parameter(default=datetime.date.today())

    def requires(self):
        return AMSLHoldings(date=self.date)

    def run(self):
        with self.input().open() as handle:
            holdings = json.load(handle)

        _, stopover = tempfile.mkstemp(prefix='siskin-')

        # The property which contains the URI of the holding file. Might change.
        urikey = 'HoldingFileURI'

        for holding in holdings:
            if holding["ISIL"] == self.isil:

                if urikey not in holding:
                    raise RuntimeError('possible AMSL API change, expected: %s, available keys: %s' % (urikey, holding.keys()))

                # refs. #7142
                if 'kbart' not in holding[urikey].lower():
                    self.logger.debug("skipping non-KBART holding URI: %s" % holding[urikey])
                    continue

                link = "%s%s" % (config.get('amsl', 'uri-download-prefix'), holding[urikey])
                downloaded = shellout("curl --fail {link} > {output} ", link=link)
                try:
                    _ = zipfile.ZipFile(downloaded)
                    output = shellout("unzip -p {input} >> {output}", input=downloaded, output=stopover)
                except zipfile.BadZipfile:
                    # at least the file is not a zip.
                    output = shellout("cat {input} >> {output}", input=downloaded, output=stopover)

        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
SWB Open Data.
"""

from gluish.format import TSV
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.database import sqlitedb
from siskin.task import DefaultTask
from siskin.utils import copyregions
import BeautifulSoup
import collections
import datetime
import luigi
import operator
import re
import requests
import string
import tempfile
import urlparse

class SWBOpenDataTask(DefaultTask):
    TAG = '025'

    def closest(self):
        if not hasattr(self, 'date'):
            raise RuntimeError('no date attribute on %s' % self)
        task = SWBOpenDataLatestDate(date=self.date)
        luigi.build([task], local_scheduler=True)
        with task.output().open() as handle:
            s = handle.read().strip()
            date = datetime.date(*map(int, s.split('-')))
        return date

class SWBOpenDataLatestDate(SWBOpenDataTask):
    """ For a given date, return the closest date in the past
    on which EBL shipped. """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return SWBOpenDataInventory()

    @timed
    def run(self):
        """ load dates and paths in pathmap, use a custom key
        function to find the closest date, but double check,
        if that date actually lies in the future - if it does, raise
        an exception, otherwise dump the closest date and filename to output """
        dates = set()
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('date', 'type', 'url')):
                date = datetime.date(*map(int, row.date.split('-')))
                dates.add(date)

        def closest_keyfun(date):
            if date > self.date:
                return datetime.timedelta(999999999)
            return abs(date - self.date)

        closest = min(dates, key=closest_keyfun)
        if closest > self.date:
            raise RuntimeError('No shipment before: %s' % min(dates))

        with self.output().open('w') as output:
            output.write_tsv(closest)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class SWBOpenDataInventory(SWBOpenDataTask):
    base = luigi.Parameter(default="http://swblod.bsz-bw.de/od/", significant=False)
    date = luigi.DateParameter(default=datetime.date.today())

    @timed
    def run(self):
        r = requests.get(self.base)
        if not r.status_code == 200:
            raise RuntimeError("HTTP %s: %s" % (r.status_code, r.text))
        soup = BeautifulSoup.BeautifulSoup(r.text)
        links = soup.findAll('a')
        packages = collections.defaultdict(set)
        patterns = {'update': r'od-up_bsz-tit_([\d]{6})_([\d]{1,}).xml.tar.gz',
                    'dump': r'od_bsz-tit_([\d]{6})_([\d]{1,}).xml.tar.gz'}
        for link in links:
            filename = link.get('href')
            kind, pattern = None, None
            if filename.startswith('od-up_bsz'):
                pattern = re.compile(patterns.get('update'))
                kind = 'update'
            elif filename.startswith('od_bsz'):
                pattern = re.compile(patterns.get('dump'))
                kind = 'dump'
            else:
                continue

            match = pattern.match(filename)
            if not match:
                raise RuntimeError("unexpected filename: %s" % filename)

            date = datetime.datetime.strptime(match.group(1), '%y%m%d').date()
            packages[date].add((kind, filename))

        with self.output().open('w') as output:
            for date in sorted(packages):
                for kind, filename in sorted(packages[date]):
                    url = urlparse.urljoin(self.base, filename)
                    output.write_tsv(date, kind, url)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class SWBOpenDataMarc(SWBOpenDataTask):
    """ Get the MARC version of SWB OD for some date. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SWBOpenDataInventory()

    @timed
    def run(self):
        """ Assumes every tarball contains a single file. Will fail during
        XML -> MARC conversion otherwise. """
        urls = set()
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('date', 'kind', 'url')):
                dateobj = datetime.date(*map(int, row.date.split('-')))
                if dateobj == self.closest():
                    urls.add(row.url)

        _, combined = tempfile.mkstemp(prefix='tasktree-')
        for url in urls:
            output = shellout("wget --retry-connrefused -O {output} '{url}'", url=url)
            output = shellout("tar -Oxzf {input} > {output}", input=output)
            shellout("yaz-marcdump -i marcxml -o marc {input} >> {output}", input=output, output=combined)
        luigi.File(combined).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class SWBOpenDataSeekMapDB(SWBOpenDataTask):
    date = ClosestDateParameter(default=datetime.date.today())
    def requires(self):
        return SWBOpenDataMarc(date=self.date)

    @timed
    def run(self):
        output = shellout("marcmap -o {output} {input}", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='db'))

class SWBOpenDataListified(SWBOpenDataTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SWBOpenDataMarc(date=self.date)

    @timed
    def run(self):
        output = shellout("""
            marctotsv -f NA -s "|" {input} 001 {date} 005 924.b > {output}""",
            date=self.closest(), input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class SWBOpenDataListifiedRange(SWBOpenDataTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        task = SWBOpenDataDates(date=self.date)
        luigi.build([task], local_scheduler=True)
        with task.output().open() as handle:
            for row in handle.iter_tsv(cols=('date',)):
                dateobj = datetime.date(*map(int, row.date.split('-')))
                yield SWBOpenDataListified(date=dateobj)

    @timed
    def run(self):
        _, combined = tempfile.mkstemp(prefix='tasktree')
        for target in self.input():
            shellout("cat {input} >> {output}", input=target.path, output=combined)
        output = shellout("LANG=C sort -k1,1 -k2,2 {input} > {output}", input=combined)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

class SWBOpenDataDates(SWBOpenDataTask):
    """ Just all the dates up to the last dump. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SWBOpenDataInventory()

    @timed
    def run(self):
        dates = set()
        with self.input().open() as handle:
            sequence = list(handle.iter_tsv(cols=('date', 'type', 'url')))
            for row in reversed(sequence):
                dateobj = datetime.date(*map(int, row.date.split('-')))
                if dateobj > self.date:
                    continue
                dates.add(row.date)
                if row.type == 'dump':
                    break
        if not dates:
            raise RuntimeError('no shipments before %s' % self.date)

        with self.output().open('w') as output:
            for date in sorted(dates):
                output.write_tsv(date)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class SWBOpenDataSnapshot(SWBOpenDataTask):
    """ Find the latest entries from SWB OD. Non-pandas version. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SWBOpenDataListifiedRange(date=self.date)

    @timed
    def run(self):
        output = shellout("""tac "{input}" | LANG=C uniq -w 9 | LANG=C cut -f 1,2 | LANG=C sort > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class SWBOpenDataSnapshotMarcX(SWBOpenDataTask):
    """ Try marcsnapshot for snapshots. High memory overhead for now. """
    date = ClosestDateParameter(default=datetime.date.today())
    limit = luigi.IntParameter(default=50000, significant=False, description='limit for seekmapdb queries')

    def requires(self):
        prerequisite = SWBOpenDataDates(date=self.date)
        luigi.build([prerequisite])
        with prerequisite.output().open() as handle:
            for row in handle.iter_tsv(cols=('date',)):
                yield SWBOpenDataMarc(date=row.date)

    def run(self):
        files = [target.path for target in self.input()]
        output = shellout("""marcsnapshot -verbose -o {output} {files}""", files=' '.join(files))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class SWBOpenDataSnapshotMarc(SWBOpenDataTask):
    """ Create a single MARC file, that represents
    the current state of affairs. """
    date = ClosestDateParameter(default=datetime.date.today())
    limit = luigi.IntParameter(default=50000, significant=False, description='limit for seekmapdb queries')

    def requires(self):
        return SWBOpenDataSnapshot(date=self.date)

    @timed
    def run(self):
        output = shellout("cut -f2 {input} | LANG=C sort | LANG=C uniq > {output}", input=self.input().path)
        with open(output) as handle:
            dates = map(string.strip, handle.readlines())

        with self.output().open('w') as output:
            for date in dates:
                dateobj = datetime.date(*map(int, date.split('-')))
                marc = SWBOpenDataMarc(date=dateobj)
                sdb = SWBOpenDataSeekMapDB(date=dateobj)
                luigi.build([marc, sdb], local_scheduler=True)
                with open(marc.output().path) as handle:
                    with sqlitedb(sdb.output().path) as cursor:
                        idset = df[df.date == date].id.values.tolist()
                        limit, offset = self.limit, 0
                        while True:
                            cursor.execute("""
                                SELECT offset, length
                                FROM seekmap WHERE id IN (%s)""" % (
                                    ','.join(("'%s'" % id for id in idset[offset:offset + limit]))))
                            rows = cursor.fetchall()
                            if not rows:
                                break
                            else:
                                copyregions(handle, output, rows)
                                offset += limit

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class SWBOpenDataSnapshotJson(SWBOpenDataTask):
    """ Create a single JSON file, that represents
    the current state of affairs. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SWBOpenDataSnapshotMarc(date=self.date)

    @timed
    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}", input=self.input().path, date=self.closest())
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

class SWBOpenDataIndex(SWBOpenDataTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'swbod'
    doc_type = 'title'
    purge_existing_index = True

    mapping = {'title': {'date_detection': False,
                          '_id': {'path': 'content.001'},
                          '_all': {'enabled': True,
                                   'term_vector': 'with_positions_offsets',
                                   'store': True}}}

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return SWBOpenDataSnapshotJson(date=self.date)

class SWBOpenDataRVK(SWBOpenDataTask):
    """ Find the RVK categories per PPN. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SWBOpenDataSnapshotMarc(date=self.date)

    def run(self):
        output = shellout("""marctotsv -s "|" {input} 001 936.a > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class SWBOpenDataRVKModel(SWBOpenDataTask):
    """ Find the coocurrences of RVK categories. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SWBOpenDataRVK(date=self.date)

    def run(self):
        cooc = collections.defaultdict(list)
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('ppn', 'rvks')):
                rvks = row.rvks.split('|')
                for key in rvks:
                    for rvk in rvks:
                        if not rvk == key:
                            cooc[key].append(rvk)

        with self.output().open('w') as output:
            for k, v in cooc.iteritems():
                output.write_tsv(k, '|'.join(v))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class SWBOpenDataRVKTopCooccurences(SWBOpenDataTask):
    """ Report the top N coocurences only. Optionally
    only take into account cooccurences set sizes greater or equal `size`.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    top = luigi.IntParameter(default=5)
    size = luigi.IntParameter(default=5)

    def requires(self):
        return SWBOpenDataRVKModel(date=self.date)

    def run(self):
        with self.input().open() as handle:
            with self.output().open('w') as output:
                for row in handle.iter_tsv(cols=('rvk', 'rvks')):
                    rvks = row.rvks.split('|')
                    total = len(rvks)
                    if total == 0:
                        continue
                    counter = collections.defaultdict(int)
                    for rvk in rvks:
                        counter[rvk] += 1
                    if len(counter.keys()) < self.size:
                        continue
                    topn = sorted([(k, v) for k, v in counter.iteritems()], key=operator.itemgetter(1), reverse=True)[:self.top]
                    topn = map(operator.itemgetter(0), topn)
                    output.write_tsv(row.rvk, *topn)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
HS ZI/GR

Configuration keys:

scp-src = [user@server:[port]]/path/to/file/{tag}/*mrc
"""

from siskin.benchmark import timed
from luigi.contrib.esindex import CopyToIndex
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.configuration import Config
from siskin.task import DefaultTask
import datetime
import luigi

config = Config.instance()

class HSZIGRTask(DefaultTask):
    TAG = '023'

    def closest(self):
        """ One time thing. """
        return datetime.date(2014, 1, 1)

class HSZIGRImport(HSZIGRTask):
    date = ClosestDateParameter(default=datetime.date.today())

    @timed
    def run(self):
        output = shellout("scp {origin} {output}", origin=config.get('hszigr', 'scp-src'))
        output = shellout("marcuniq -x 'BW, ZI' {input} > {output}", input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class HSZIGRJson(HSZIGRTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return HSZIGRImport(date=self.date)

    @timed
    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                          input=self.input().path, date=self.closest())
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class HSZIGRIndex(HSZIGRTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'hszigr'
    doc_type = 'title'
    pruge_existing_index = True

    mapping = {'title': {'date_detection': False,
                      '_id': {'path': 'content.001'},
                      '_all': {'enabled': True,
                               'term_vector': 'with_positions_offsets',
                               'store': True}}}
    def update_id(self):
        return self.effective_task_id()

    def requires(self):
        return HSZIGRJson(date=self.date)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

from siskin.benchmark import timed
from gluish.common import Executable
from gluish.format import TSV
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.utils import memoize
from siskin.task import DefaultTask
import BeautifulSoup
import datetime
import luigi
import re

class WikidataTask(DefaultTask):
    """
    Base task for wikidata.
    """
    TAG = 'wikidata'

    @memoize
    def closest(self):
        if not hasattr(self, 'date'):
            raise RuntimeError('Task has no date: %s' % self)
        task = WikidataLatestDate(date=self.date)
        luigi.build([task], local_scheduler=True)
        with task.output().open() as handle:
            date = handle.iter_tsv(cols=('date', 'url')).next().date
        result = datetime.date(*(int(v) for v in date.split('-')))
        return result

class WikidataLatestDate(WikidataTask):
    """
    For a given date, return the closest date in the past on which wikidata shipped.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return WikidataDumpDates()

    @timed
    def run(self):
        """ load dates and urls in urlmap, use a custom key
        function to find the closest date, but double check,
        if that date actually lies in the future - if it does, raise
        an exception, otherwise dump the closest date and filename to output """
        urlmap = {}
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('date', 'url')):
                date = datetime.datetime.strptime(row.date, '%Y-%m-%d').date()
                urlmap[date] = row.url

        def closest_keyfun(date):
            if date > self.date:
                return datetime.timedelta(999999999)
            return abs(date - self.date)

        closest = min(urlmap.keys(), key=closest_keyfun)
        if closest > self.date:
            raise RuntimeError('No shipment before: %s' % min(urlmap.keys()))

        with self.output().open('w') as output:
            output.write_tsv(closest, urlmap.get(closest))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WikidataDumpDates(WikidataTask):
    """ Dump dates and links like this:

    2014-02-10  http://dumps.wikimedia.org/wikidatawiki/20140210/
    2014-02-26  http://dumps.wikimedia.org/wikidatawiki/20140226/
    2014-03-15  http://dumps.wikimedia.org/wikidatawiki/20140315/
    ...
    """
    date = luigi.DateParameter(default=datetime.date.today())

    @timed
    def run(self):
        output = shellout("""wget --retry-connrefused http://dumps.wikimedia.org/wikidatawiki/ -O {output}""")
        with open(output) as handle:
            soup = BeautifulSoup.BeautifulSoup(handle.read())

        pattern = re.compile(r".*?(\d{8,8}).*?")
        dates = []
        for link in soup.findAll('a'):
            if pattern.match(link.text):
                date = datetime.date(int(link.text[0:4]), int(link.text[4:6]),
                                     int(link.text[6:8]))
                url = "http://dumps.wikimedia.org/wikidatawiki/%s" % link['href']
                dates.append((date, url))

        with self.output().open('w') as output:
            for date, url in sorted(dates):
                output.write_tsv(date, url)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WikidataDumpLinks(WikidataTask):
    """ A wikidata dump will consist of many files. Dump all links for those files. """

    date = luigi.DateParameter(default=datetime.date.today())

    @timed
    def run(self):
        url = "http://dumps.wikimedia.org/wikidatawiki/%s/" % (self.closest().strftime('%Y%m%d'))
        output = shellout("""wget --retry-connrefused {url} -O {output}""", url=url)

        with open(output) as handle:
            soup = BeautifulSoup.BeautifulSoup(handle.read())

        pattern = re.compile("wikidatawiki-.*")
        links = set()
        for link in soup.findAll('a'):
            if pattern.match(link.text):
                date = self.closest().strftime('%Y%m%d')
                url = "http://dumps.wikimedia.org/wikidatawiki/%s/%s" % (date, link.text)
                links.add(url)

        with self.output().open('w') as output:
            for url in sorted(links):
                output.write_tsv(url)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WikidataFile(WikidataTask):
    """
    Download a single file with a certain suffix out of the dump files.
    The most important is pages-articles.xml.bz2, which is the default.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    suffix = luigi.Parameter(default='pages-articles.xml.bz2')

    timeout = luigi.IntParameter(default=10, significant=False)

    def requires(self):
        return WikidataDumpLinks(date=self.date)

    @timed
    def run(self):
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('url',)):
                if row.url.endswith(self.suffix):
                    output = shellout("""curl --connect-timeout {timeout} {url} | bunzip2 -c > {output}""",
                                      url=row.url, timeout=self.timeout)
                    luigi.File(output).move(self.output().path)
                    break
            else:
                raise RuntimeError('No URL with suffix: %s, try to rerun WikidataDumpLinks' % self.suffix)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True, ext='xml'))

class WikiPagesJson(WikidataTask):
    """
    Generic conversion from XML to JSON.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'dump': WikidataFile(suffix='pages-articles.xml.bz2', date=self.date),
                'app': Executable(name='wikidatatojson', message='https://github.com/miku/wikitools')}

    @timed
    def run(self):
        output = shellout("wikidatatojson {input} > {output}", input=self.input().get('dump').path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True, ext='ldj'))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.task import DefaultTask
import datetime
import luigi
import tempfile

class DissonTask(DefaultTask):
    TAG = '013'

    def closest(self):
        return monthly(date=self.date)

class DissonHarvest(luigi.WrapperTask, DissonTask):
    """ Harvest Disson. """

    begin = luigi.DateParameter(default=datetime.date(2000, 1, 1))
    end = ClosestDateParameter(default=datetime.date.today())
    prefix = luigi.Parameter(default="oai_dc", significant=False)
    url = luigi.Parameter(default="http://services.dnb.de/oai/repository", significant=False)
    collection = luigi.Parameter(default="dnb:online:dissertations", significant=False)

    def requires(self):
        """ Harvest in 1-month chunks. """
        if self.end < self.begin:
            raise RuntimeError('Invalid range: %s - %s' % (self.begin, self.end))
        dates = date_range(self.begin, self.end, 1, 'months')
        for begin, end in zip(dates[:-1], dates[1:]):
            yield OAIHarvestChunk(begin=begin, end=end, prefix=self.prefix,
                                  url=self.url, collection=self.collection)

    def output(self):
        return self.input()

class DissonCombine(DissonTask):
    """ Combine the chunks for a date range into a single file.
    The filename will carry a name, that will only include year
    and month for begin and end, example:

    .../013/disson-combine/2010-01--2013-11.mrc"""

    begin = luigi.DateParameter(default=datetime.date(2000, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())
    prefix = luigi.Parameter(default="oai_dc", significant=False)
    url = luigi.Parameter(default="http://services.dnb.de/oai/repository", significant=False)
    collection = luigi.Parameter(default="dnb:online:dissertations", significant=False)

    def requires(self):
        return {'files': DissonHarvest(begin=self.begin, end=self.date,
                                       prefix=self.prefix, url=self.url,
                                       collection=self.collection),
                'apps': [Executable(name='xsltproc'),
                         Executable(name='yaz-marcdump'),
                         Executable(name='marcuniq')]}

    @timed
    def run(self):
        xsl = self.assets("013_OAIDCtoMARCXML.xsl")
        _, combined = tempfile.mkstemp(prefix='siskin-')
        for target in self.input().get('files'):
            output = shellout("xsltproc {stylesheet} {input} > {output}",
                              stylesheet=xsl, input=target.path)
            output = shellout("yaz-marcdump -i marcxml -o marc {input} > {output}",
                              input=output)
            shellout("cat {input} >> {output}", input=output, output=combined)

        output = shellout("marcuniq -o {output} {input}", input=combined)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class DissonJson(DissonTask):
    """ Convert to JSON. """

    begin = luigi.DateParameter(default=datetime.date(2000, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return DissonCombine(begin=self.begin, date=self.date)

    def run(self):
        tmp = shellout("marctojson -m date={date} {input} > {output}",
                       input=self.input().path, date=self.date)
        luigi.File(tmp).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class DissonIndex(DissonTask, CopyToIndex):
    begin = luigi.DateParameter(default=datetime.date(2000, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    index = "disson"
    doc_type = "title"
    purge_existing_index = True

    mapping = {'title': {'date_detection': False,
                          '_id': {'path': 'content.001'},
                          '_all': {'enabled': True,
                                   'term_vector': 'with_positions_offsets',
                                   'store': True}}}

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return DissonJson(begin=self.begin, date=self.date)
